#include "spi.h"

spi::SPI::SPI(void){}

spi::SPI::~SPI(void){}

void spi::SPI::Init(SPI_TypeDef *Instance)
{
        this->Instance = Instance;
        //Now we start with the SPI Peripheral
        //We have the data capture on the second edge, are the master and are REALLY SLOW and SS is done in Software
        this->Instance->CR1 = SPI_CR1_MSTR | SPI_CR1_BR | SPI_CR1_SSM | SPI_CR1_SSI;
        //Our data size is 8 bytes and the recieve fifo not empty status bit is activated as soon as there is any data
        //(makes it easier to handle interrupt based transfers with an uneven number of bytes)
        this->Instance->CR2 = SPI_CR2_DS_0 | SPI_CR2_DS_1 | SPI_CR2_DS_2 | SPI_CR2_FRXTH;
        this->Instance->CR1 |= SPI_CR1_SPE;
}

uint8_t spi::SPI::TradeByte(uint8_t transmit)
{
        *((uint8_t *)&(Instance->DR)) = transmit;
        //Wait until a byte is available
        while(!(Instance->SR & SPI_SR_FRLVL));
        return *((uint8_t *)&(Instance->DR));

}

void spi::SPI::InterruptHandler(void)
{
        if(Instance->SR & SPI_SR_TXE){
                while(((Instance->SR & SPI_SR_FTLVL) != SPI_SR_FTLVL) && bufferOutPos < bufferLength){
                        *((uint8_t *)&Instance->DR) = bufferOut[bufferOutPos++];
                }
                //Wenn keine TX Daten mehr vorliegen sollten wir sie deaktivieren, sonst haben wir die ganze Zeit Interrupts
        }
        if(Instance->SR & SPI_SR_RXNE){
                while((Instance->SR & SPI_SR_FRLVL) && bufferInPos < bufferLength){
                                bufferIn[bufferInPos++] = *((uint8_t *)&Instance->DR);
                }
        }
        if((bufferInPos == bufferLength) && (bufferOutPos == bufferLength)){
                Instance->CR2 &= ~(SPI_CR2_RXNEIE | SPI_CR2_TXEIE);
                bufferLength = 0;
                d->TransferEndCallback();
        }

}

uint8_t spi::SPI::StartTransfer(uint8_t *out, uint8_t *in, uint8_t length, Device *callingDevice)
{
        if(bufferLength != 0)
                return 1;
        bufferOut = out;
        bufferIn = in;
        bufferLength = length;
        bufferInPos = 0;
        bufferOutPos = 0;
        d = callingDevice;
        //First lets clear the RX buffer (just to be sure)
        while(Instance->SR & SPI_SR_FRLVL){
                //This should do a read access (as it is declared volatile)
                //and just throw away the data
                uint8_t x = (uint8_t) Instance->DR;
                //So there are no unused variable errors
                x++;
        }
        //Now start the interrupt
        Instance->CR2 |= SPI_CR2_RXNEIE | SPI_CR2_TXEIE;
        return 0;
}

uint8_t spi::SPI::GetTransferStatus(void)
{
        if(bufferLength == 0)
                return 0;
        else
                return 1;
}


spi::Device::Device(void){}
spi::Device::~Device(void){}

void spi::Device::Init(SPI &SPIInstance, gpio::pins pin)
{
        this->pin = pin;
        gpio::mode(this->pin, gpio::outputPP);
        this->SPIInstance = &SPIInstance;
        Deselect();
}

uint8_t spi::Device::StartTransfer(void (*transferCallback)(void),uint8_t *sendData, uint8_t length)
{
        if((status & 1)== 1)
                return 2;
        if(SPIInstance->GetTransferStatus())
                return 1;
        if(length > SPI_BUFFER_LENGTH)
                return 3;
        status |= 1;
        this->transferCallback = transferCallback;
        this->transferLength = length;
        for(uint8_t i = length; i>0; i--)
                this->bufferOut[i-1] = *((sendData+i)-1);
        Select();
        return this->SPIInstance->StartTransfer(bufferOut, bufferIn, this->transferLength, this);
}

uint8_t spi::Device::TradeByte(uint8_t transmit)
{
        uint8_t receive;
        receive =  SPIInstance->TradeByte(transmit);
        return receive;
}

void spi::Device::Select(void)
{
        gpio::low(pin);
}

void spi::Device::Deselect(void)
{
        gpio::high(pin);
}

void spi::Device::TransferEndCallback(void)
{
        status &= ~(0b1);
        status |= 0b10;
        Deselect();
        transferCallback();
}

uint8_t spi::Device::GetData(uint8_t *data)
{
        for(uint8_t i = transferLength; i>0; i--)
                *((data+i)-1) = this->bufferIn[i-1];
        status &= ~(0b10);
        uint8_t length = transferLength;
        transferLength = 0;
        return length;
}

uint8_t spi::Device::GetTransferStatus(void)
{
        switch(status){
        case 0: return 0;
                break;
        case 0b1: return 1;
                break;
        case 0b10: return 2;
                break;
        default: return 0xff;
                break;
        }
}

void spi::specificInitDefault(void)
{
        //Enable Clocks
        RCC->AHBENR |= RCC_AHBENR_GPIOAEN;
        RCC->APB2ENR |= RCC_APB2ENR_SPI1EN;
        //Reset all the Pins we need to input mode
        GPIOA->MODER &= ~(GPIO_MODER_MODER5 | GPIO_MODER_MODER6 | GPIO_MODER_MODER7);
        //GPIO4 in Ouput, the rest as AF
        GPIOA->MODER |= GPIO_MODER_MODER5_1 | GPIO_MODER_MODER6_1 | GPIO_MODER_MODER7_1;
        //We want the Pins to be in push-pull mode
        GPIOA->OTYPER &= ~(GPIO_OTYPER_OT_5 | GPIO_OTYPER_OT_6 | GPIO_OTYPER_OT_7);
        //We want them to be in High Speed mode (since this IS SPI)
        GPIOA->OSPEEDR |= GPIO_OSPEEDR_OSPEEDR5 | GPIO_OSPEEDR_OSPEEDR6 | GPIO_OSPEEDR_OSPEEDR7;
        //The we disable all pull-ups/downs
        GPIOA->PUPDR |= 0 | GPIO_PUPDR_PUPDR5_1 | GPIO_PUPDR_PUPDR6_1 | GPIO_PUPDR_PUPDR7_1;
        //Then we choose SPI as the Af
        GPIOA->AFR[0] &= ~(GPIO_AFRL_AFRL5 | GPIO_AFRL_AFRL6 | GPIO_AFRL_AFRL7);
}
